
def fix_scaling(feature):
    """
    对数据框中的数据进行中性化处理

    Args:
        feature: 待处理的数据框

    Returns:
        feature: 已处理数据框
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    for c in numeric_var:
        sm = feature[c].describe()
        feature[c] = feature[c].apply(lambda x: (x - sm['mean'])/sm['std'] if x else x)
    return feature



from math import log
import pandas as pd

def make_card(coefs, woe_result, odds, score, pdo):
    # odds = P_0 / P_1
    b = pdo / log(2)
    a = score - b * log(odds)
    score_card = pd.DataFrame([['base_score', '-', '-', int(a - b * coefs['intercept_'])]],
                              columns=['Variables', 'Bins', 'Woe', 'Score'])
    for v in coefs:
        if v != 'intercept_':
            woe_result[v].insert(loc=0, column='Variables', value=v)
            woe_result[v].rename(columns={v: "Bins", 'woe_i': 'Woe'}, inplace=True)
            woe_result[v]['Score'] = woe_result[v]['Woe'].apply(lambda x: int(-x * b * coefs[v]))
            score_card = pd.concat([score_card, woe_result[v]])
    score_card.to_excel('result/score_card.xlsx', index=False)
    return score_card




def fix_missing_value(feature, threshold=0.05):
    """
    修正数据框中的数值型变量中的缺失值

    Args:
        feature: 待修正的数据框
        threshold: 缺失值替换阈值

    Returns:
        feature: 已处理数据框
        missing_filling: 缺失值统计结果
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    missing_filling = {'result': {}, 'replace': []}
    for c in feature:
        missing_rate = feature[c].isna().sum()/len(feature)
        if missing_rate <= threshold and missing_rate:
            missing_filling['result'][c] = {'fill_number': feature[c].mean(), 'missing_value_rate': missing_rate}
            missing_filling['replace'].append(c)
            if c in numeric_var:
                feature[c].fillna(missing_filling['result'][c], inplace=True)
            else:
                feature[c].fillna(feature[c].mode(), inplace=True)
    return feature, missing_filling



import pickle
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import auc, roc_curve, accuracy_score, recall_score
import matplotlib.pyplot as plt
import mksc
from mksc.feature_engineering import preprocess
from .custom import Custom
from mksc.feature_engineering import FeatureEngineering

def training(x_train, y_train, x_test, y_test, x_valid, y_valid):
    """
    模型训练过程函数
    1. 训练
    2. 预测
    3. 评估
    4. TODO 模型可选

    Args:
        x_train: 训练集特征
        y_train: 训练集标签
        x_test: 测试集特征
        y_test: 测试集标签
        x_valid: 验证集特征
        y_valid: 验证集标签
    """
    model = LogisticRegression()
    model.fit(x_train, y_train)

    # 预测结果
    predict_train = model.predict(x_train)
    predict_valid = model.predict(x_valid)
    predict_test = model.predict(x_test)

    # 模型评估
    acu_train = accuracy_score(y_train, predict_train)
    acu_valid = accuracy_score(y_valid, predict_valid)
    acu_test = accuracy_score(y_test, predict_test)

    sen_train = recall_score(y_train, predict_train, pos_label=1)
    sen_valid = recall_score(y_valid, predict_valid, pos_label=1)
    sen_test = recall_score(y_test, predict_test, pos_label=1)

    spe_train = recall_score(y_train, predict_train, pos_label=0)
    spe_valid = recall_score(y_valid, predict_valid, pos_label=0)
    spe_test = recall_score(y_test, predict_test, pos_label=0)
    print(f'模型准确率：验证 {acu_valid * 100:.2f}%	训练 {acu_train * 100:.2f}%	测试 {acu_test * 100:.2f}%')
    print(f'正例覆盖率：验证 {sen_valid * 100:.2f}%	训练 {sen_train * 100:.2f}%	测试 {sen_test * 100:.2f}%')
    print(f'负例覆盖率：验证 {spe_valid * 100:.2f}%	训练 {spe_train * 100:.2f}%	测试 {spe_test * 100:.2f}%')

    # K-s & roc
    predict_train_prob = np.array([i[1] for i in model.predict_proba(x_train)])
    fpr, tpr, thresholds = roc_curve(y_train.values, predict_train_prob, pos_label=1)
    auc_score = auc(fpr, tpr)
    w = tpr - fpr
    ks_score = w.max()
    ks_x = fpr[w.argmax()]
    ks_y = tpr[w.argmax()]
    fig, ax = plt.subplots()
    ax.plot(fpr, tpr, label='AUC=%.5f' % auc_score)
    ax.set_title('Receiver Operating Characteristic')
    ax.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6))
    ax.plot([ks_x, ks_x], [ks_x, ks_y], '--', color='red')
    ax.text(ks_x, (ks_x + ks_y) / 2, '  KS=%.5f' % ks_score)
    ax.legend()
    fig.savefig("result/ks_roc.png")

    # 模型保存
    coefs = dict(list(zip(x_train.columns, list(model.coef_[0]))) + [("intercept_", model.intercept_[0])])
    with open('result/model.pickle', 'wb') as f:
        f.write(pickle.dumps(model))
    with open('result/coefs.pickle', 'wb') as f:
        f.write(pickle.dumps(coefs))


def main():
    """
    项目训练程序入口
    """
    # 加载数据、变量类型划分、特征集与标签列划分
    data = mksc.load_data()
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = data[label_var]

    # 自定义数据清洗
    feature, label = Custom.clean_data(feature, label)

    # 数据类型转换
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')

    # One-Hot编码
    # feature = pd.concat([feature, pd.get_dummies(feature[category_var])], axis=1)
    # feature.drop(category_var+datetime_var, inplace=True)

    # 特征工程
    # 自定义特征组合模块
    feature = Custom.feature_combination(feature)

    fe = FeatureEngineering(feature, label)
    feature = fe.run()

    # 数据集划分
    x_train, x_valid_test, y_train, y_valid_test = train_test_split(feature, label, test_size=0.4, random_state=0)
    x_valid, x_test, y_valid, y_test = train_test_split(x_valid_test, y_valid_test, test_size=0.5, random_state=0)

    # 模型训练
    training(x_train, y_train, x_test, y_test, x_valid, y_valid)


if __name__ == "__main__":
    main()



import pickle
from mksc.feature_engineering import seletction
from mksc.feature_engineering import values
from mksc.feature_engineering import binning

class FeatureEngineering(object):
    def __init__(self, feature, label, missing_threshold=(0.9, 0.05), distinct_threshold=0.9, unique_threshold=0.9,
                 abnormal_threshold=0.05, correlation_threshold=0.7):
        self.feature = feature
        self.label = label
        self.threshold = {}
        self.missing_threshold = missing_threshold
        self.distinct_threshold = distinct_threshold
        self.unique_threshold = unique_threshold
        self.abnormal_threshold = abnormal_threshold
        self.correlation_threshold = correlation_threshold

    def run(self):
        """
        特征工程过程函数,阈值参数可以自定义修改
        1. 特征组合
        2. 基于统计特性特征选择：缺失率、唯一率、众数比例
        3. 极端值处理
        4. TODO 异常值处理
        5. 缺失值处理
        6. COMMENT 归一化处理
        7. COMMENT 正态化处理
        8. 最优分箱
        9. IV筛选
        10. TODO PSI筛选
        11. 相关性筛选
        12. woe转化
        13. 逐步回归筛选

        Returns:
            feature： 已完成特征工程的数据框
        """
        feature = self.feature
        label = self.label
        # 基于缺失率、唯一率、众数比例统计特征筛选
        missing_value = seletction.get_missing_value(feature, self.missing_threshold[0])
        distinct_value = seletction.get_distinct_value(feature, self.distinct_threshold)
        unique_value = seletction.get_unique_value(feature, self.unique_threshold)
        feature.drop(set(missing_value['drop'] + distinct_value['drop'] + unique_value['drop']), axis=1, inplace=True)

        # 极端值处理
        feature, abnormal_value = values.fix_abnormal_value(feature, self.abnormal_threshold)

        # 缺失值处理
        feature, missing_filling = values.fix_missing_value(feature, self.missing_threshold[1])

        # 归一化处理
        # feature = pp.fix_scaling(feature)

        # 正态化处理
        # feature, standard_lambda = pp.fix_standard(feature)

        # 数值特征最优分箱，未处理的变量，暂时退出模型
        bin_result, iv_result, woe_result, woe_adjust_result = binning.tree_binning(label, feature)
        bin_error_drop = bin_result['error'] + woe_adjust_result

        # IV筛选
        iv_drop = list(filter(lambda x: iv_result[x] < 0.02, iv_result))
        feature.drop(iv_drop + bin_error_drop, inplace=True, axis=1)

        # 相关性筛选
        cor_drop = seletction.get_cor_drop(feature, iv_result, self.correlation_threshold)
        feature.drop(cor_drop, inplace=True, axis=1)

        # woe转化
        feature = binning.woe_transform(feature, woe_result, bin_result)

        # 逐步回归筛选
        feature_selected = seletction.stepwise_selection(feature, label)
        feature = feature[feature_selected]

        # 中间结果保存
        result = {"missing_value": missing_value,
                  "distinct_value": distinct_value,
                  "unique_value": unique_value,
                  "abnormal_value": abnormal_value,
                  "missing_filling": missing_filling,
                  "bin_result": bin_result,
                  "iv_result": iv_result,
                  "woe_result": woe_result,
                  "woe_adjust_result": woe_adjust_result,
                  "bin_error_drop": bin_error_drop,
                  "iv_drop": iv_drop,
                  "cor_drop": cor_drop,
                  "feature_selected": feature_selected
                  }
        with open('result/feature_engineering.pickle', 'wb') as f:
            f.write(pickle.dumps(result))
        return feature



from statsmodels.iolib.smpickle import load_pickle
from feature_engineering.scoring import process as ps
import configparser
import os

def main():
    """
    评分卡制作主程序入口
    """
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(os.getcwd(), 'conf', 'config/configuration.ini'), encoding='utf_8_sig')
    odds = cfg.get('configuration', 'odds')
    score = cfg.get('configuration', 'score')
    pdo = cfg.get('configuration', 'pdo')
    woe_result = load_pickle("result/feature_engineering.pickle")["woe_result"]
    coefs = load_pickle("result/coefs.pickle")
    ps.make_card(coefs, woe_result, odds, score, pdo)


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
import re

class Custom(object):
    """
    自定义预处理类函数封装
    """
    @staticmethod
    def clean_data(feature, label):
        """
        基于单元格与行操作的数据清洗

        Args:
            feature: 待清洗特征数据框
            label: 待清洗标签序列

        Returns:
            feature_tmp: 已清洗特征数据框
            label_tmp: 已清洗标签序列
        """
        feature_tmp = feature.copy()
        label_tmp = label.copy()
        # 默认值处理
        # 单位处理
        # 正则处理
        # 业务缺失值补充
        # 业务异常值剔除
        return feature_tmp, label_tmp

    @staticmethod
    def feature_combination(feature):
        """
        基于列操作的数据清洗与特征构造

        Args:
            feature: 待清洗特征数据框

        Returns:
            feature_tmp: 已清洗特征数据框
        """
        feature_tmp = feature.copy()
        # 构造衍生变量
        # 日期变量处理
        return feature_tmp

    @staticmethod
    def custom_model(feature, label):
        """
        TODO 自定义模型
        """
        model = (feature, label)
        return model
